/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * transport_sock.c
 * ----------------
 *
 * Communication transport functions
 * Sockets protocols
 * 
 * Addresses are specified in the form:
 * 
 * protocol:node:service
 * 
 * For IP based services, 'node' can be a host name,
 * IPv4 dotted decimal address, or IPv6 colon separated
 * hexidecimal address. 'service' is either a numeric
 * port number or service name from /etc/services or the like
 * 
 * For Unix domain sockets, 'node' must be one of 'unix' or
 * 'local', and 'service' is a full path name.
 * 
 * eg.
 * 
 * sock:www.google.com:http
 * sock:10.0.0.1:ftp
 * sock:localhost:6000
 * sock:unix:/var/run/mysocket
 * sock:f340:f353::f454:telnet
 * 
 */

#include "../../../config.h"

#include "impulse/comms/transport_sock_int.h"
#include "impulse/comms/transport_int.h"

#include "impulse/util/memory.h"
#include "impulse/util/error.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#if HAVE_COMMS_SOCK

#include <netdb.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/ioctl.h>

/*                            ======--------======                           */



union sockaddr_all
{
  struct sockaddr a;
#if HAVE_COMMS_SOCK_UNIX || HAVE_COMMS_SOCK_LOCAL
  struct sockaddr_un un;
#endif
#if HAVE_COMMS_SOCK_IPV4
  struct sockaddr_in in;
#endif
#if HAVE_COMMS_SOCK_IPV6
  struct sockaddr_in6 in6;
#endif
};


typedef struct _impulse_transport_sock_info
{
  impulse_int family;
  impulse_int protocol;
  impulse_int type;
  impulse_unistring name;
} impulse_transport_sock_info;



/*                            ======--------======                           */


/* Static unicode strings for protocol name */
static struct _impulse_unistring _name_tcp =
{
  3, 3,
  (impulse_unichar[]){
    (impulse_unichar)'t',
    (impulse_unichar)'c',
    (impulse_unichar)'p', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_udp =
{
  3, 3,
  (impulse_unichar[]){
    (impulse_unichar)'u',
    (impulse_unichar)'d',
    (impulse_unichar)'p', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_tcp4 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'t',
    (impulse_unichar)'c',
    (impulse_unichar)'p', 
    (impulse_unichar)'4', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_udp4 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'u',
    (impulse_unichar)'d',
    (impulse_unichar)'p', 
    (impulse_unichar)'4', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_tcp6 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'t',
    (impulse_unichar)'c',
    (impulse_unichar)'p', 
    (impulse_unichar)'6', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_udp6 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'u',
    (impulse_unichar)'d',
    (impulse_unichar)'p', 
    (impulse_unichar)'6', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_tcpl =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'t',
    (impulse_unichar)'c',
    (impulse_unichar)'p', 
    (impulse_unichar)'l', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_udpl =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'u',
    (impulse_unichar)'d',
    (impulse_unichar)'p', 
    (impulse_unichar)'l', 
    (impulse_unichar)'\0'
  }
};

/* Transport function blocks */

static struct _impulse_transport_funcs _sock_funcs =
{
  impulse_transport_sock_create,
  impulse_transport_sock_destroy,

/* Event functions */
  impulse_transport_sock_handle,
  impulse_transport_sock_unhandle,

/* Connection functions */
  impulse_transport_sock_open,
  impulse_transport_sock_accept,
  impulse_transport_sock_close,
  impulse_transport_sock_active,

/* Data transfer functions */
  impulse_transport_sock_pending,
  impulse_transport_sock_send,
  impulse_transport_sock_receive,
  impulse_transport_sock_peek,

/* Address functions */
  impulse_transport_sock_local,
  impulse_transport_sock_remote,
};


struct _impulse_transport _impulse_transport_sock_tcp = 
{
  &_name_tcp,
  IMPULSE_TRANSPORT_REMOTE | IMPULSE_TRANSPORT_RELIABLE | 
    IMPULSE_TRANSPORT_SEQUENCED | IMPULSE_TRANSPORT_STREAM |
    IMPULSE_TRANSPORT_CONNECTION | IMPULSE_TRANSPORT_CLIENTSERVER,
  &_sock_funcs
};
struct _impulse_transport _impulse_transport_sock_udp = 
{
  &_name_udp,
  IMPULSE_TRANSPORT_REMOTE | IMPULSE_TRANSPORT_FAST |
    IMPULSE_TRANSPORT_QUICK  | IMPULSE_TRANSPORT_DATAGRAM |
    IMPULSE_TRANSPORT_CONNECTIONLESS | IMPULSE_TRANSPORT_PEER,
  &_sock_funcs
};

struct _impulse_transport _impulse_transport_sock_tcp4 =
{
  &_name_tcp4,
  IMPULSE_TRANSPORT_REMOTE | IMPULSE_TRANSPORT_RELIABLE | 
    IMPULSE_TRANSPORT_SEQUENCED | IMPULSE_TRANSPORT_STREAM | 
    IMPULSE_TRANSPORT_CONNECTION | IMPULSE_TRANSPORT_CLIENTSERVER,
  &_sock_funcs
};
struct _impulse_transport _impulse_transport_sock_udp4 = 
{
  &_name_udp4,
  IMPULSE_TRANSPORT_REMOTE | IMPULSE_TRANSPORT_FAST |
    IMPULSE_TRANSPORT_QUICK  | IMPULSE_TRANSPORT_DATAGRAM |
    IMPULSE_TRANSPORT_CONNECTIONLESS | IMPULSE_TRANSPORT_PEER,
  &_sock_funcs
};

struct _impulse_transport _impulse_transport_sock_tcp6 = 
{
  &_name_tcp6,
  IMPULSE_TRANSPORT_REMOTE | IMPULSE_TRANSPORT_RELIABLE | 
    IMPULSE_TRANSPORT_SEQUENCED | IMPULSE_TRANSPORT_STREAM | 
    IMPULSE_TRANSPORT_CONNECTION | IMPULSE_TRANSPORT_CLIENTSERVER,
  &_sock_funcs
};
struct _impulse_transport _impulse_transport_sock_udp6 = 
{
  &_name_udp6,
  IMPULSE_TRANSPORT_REMOTE | IMPULSE_TRANSPORT_FAST |
    IMPULSE_TRANSPORT_QUICK | IMPULSE_TRANSPORT_DATAGRAM |
    IMPULSE_TRANSPORT_CONNECTIONLESS | IMPULSE_TRANSPORT_PEER,
  &_sock_funcs
};

struct _impulse_transport _impulse_transport_sock_tcpl = 
{
  &_name_tcpl,
  IMPULSE_TRANSPORT_RELIABLE | IMPULSE_TRANSPORT_FAST |
    IMPULSE_TRANSPORT_SEQUENCED | IMPULSE_TRANSPORT_STREAM | 
    IMPULSE_TRANSPORT_CONNECTION | IMPULSE_TRANSPORT_CLIENTSERVER |
    IMPULSE_TRANSPORT_SECURE,
  &_sock_funcs
};
struct _impulse_transport _impulse_transport_sock_udpl = 
{
  &_name_udpl,
  IMPULSE_TRANSPORT_RELIABLE | IMPULSE_TRANSPORT_SEQUENCED |
    IMPULSE_TRANSPORT_FAST | IMPULSE_TRANSPORT_QUICK |
    IMPULSE_TRANSPORT_DATAGRAM | IMPULSE_TRANSPORT_CONNECTIONLESS | 
    IMPULSE_TRANSPORT_PEER | IMPULSE_TRANSPORT_SECURE, 
  &_sock_funcs
};

impulse_transport impulse_transport_sock_tcp = &_impulse_transport_sock_tcp;
impulse_transport impulse_transport_sock_udp = &_impulse_transport_sock_udp;
impulse_transport impulse_transport_sock_tcp4 = &_impulse_transport_sock_tcp4;
impulse_transport impulse_transport_sock_udp4 = &_impulse_transport_sock_udp4;
impulse_transport impulse_transport_sock_tcp6 = &_impulse_transport_sock_tcp6;
impulse_transport impulse_transport_sock_udp6 = &_impulse_transport_sock_udp6;
impulse_transport impulse_transport_sock_tcpl = &_impulse_transport_sock_tcpl;
impulse_transport impulse_transport_sock_udpl = &_impulse_transport_sock_udpl;


impulse_transport_sock_info sock_info[IMPULSE_TRANSPORT_SOCK_MAX] = 
{
  { AF_UNSPEC, PF_UNSPEC, SOCK_STREAM, &_name_tcp},
  { AF_UNSPEC, PF_UNSPEC, SOCK_DGRAM, &_name_udp},
  
#if HAVE_COMMS_SOCK_IPV4
  { AF_INET, PF_INET, SOCK_STREAM, &_name_tcp4},
  { AF_INET, PF_INET, SOCK_DGRAM, &_name_udp4},
#else
  { AF_UNSPEC, PF_UNSPEC, SOCK_STREAM, NULL},
  { AF_UNSPEC, PF_UNSPEC, SOCK_DGRAM, NULL},
#endif
  
#if HAVE_COMMS_SOCK_IPV6
  { AF_INET6, PF_INET6, SOCK_STREAM, &_name_tcp6},
  { AF_INET6, PF_INET6, SOCK_DGRAM, &_name_udp6},
#else
  { AF_UNSPEC, PF_UNSPEC, SOCK_STREAM, NULL},
  { AF_UNSPEC, PF_UNSPEC, SOCK_DGRAM, NULL},
#endif
  
#if HAVE_COMMS_SOCK_LOCAL
  { AF_LOCAL, PF_LOCAL, SOCK_STREAM, &_name_tcpl},
  { AF_LOCAL, PF_LOCAL, SOCK_DGRAM, &_name_udpl}
#elif HAVE_COMMS_SOCK_UNIX
  { AF_UNIX, PF_UNIX, SOCK_STREAM, &_name_tcpl},
  { AF_UNIX, PF_UNIX, SOCK_DGRAM, &_name_udpl}
#else
  { AF_UNSPEC, PF_UNSPEC, SOCK_STREAM, NULL},
  { AF_UNSPEC, PF_UNSPEC, SOCK_DGRAM, NULL},
#endif
};

/*                            ======--------======                           */

/* Miscellaneous helper functions */

static impulse_int impulse_transport_sock_split(impulse_unistring address,
						impulse_int *protocol,
						impulse_char **node,
						impulse_char **service)
{
  impulse_int offset1 = impulse_unistring_find_chr(address, (impulse_unichar)':');
  impulse_int offset2 = impulse_unistring_rfind_chr(address, (impulse_unichar)':');
  impulse_unistring uprotocol = NULL;
  impulse_unistring unode = NULL;
  impulse_unistring uservice = NULL;
  impulse_int i;

  *protocol = -1;
  if (*node)
    impulse_free(node);
  if (*service)
    impulse_free(service);
  *node = NULL;
  *service = NULL;
  
  if (offset1 < 0 || offset2 < 0 ||
      offset1 == offset2)
    return -1;

  uprotocol = impulse_unistring_create_space(offset1);
  impulse_unistring_substring(uprotocol, address, 0, offset1);
  for (i = 0 ; i < IMPULSE_TRANSPORT_SOCK_MAX ; i++)
  {
    if (impulse_unistring_equal(uprotocol, sock_info[i].name))
    {
      *protocol = i;
      break;
    }
  }
  impulse_unistring_destroy(uprotocol);

  if (*protocol == -1)
    return -1;
  
  /* First the node */
  unode = impulse_unistring_create_space(offset2-(offset1+1));
  impulse_unistring_substring(unode, address, offset1+1, offset2);
  *node = impulse_unistring_to_utf(unode, 8, 0);
  impulse_unistring_destroy(unode);

  /* And now the service */
  uservice = impulse_unistring_create_space(address->length - (offset2+1));
  impulse_unistring_substring(uservice, address, offset2+1, address->length);
  *service = impulse_unistring_to_utf(uservice, 8, 0);
  impulse_unistring_destroy(uservice);

  return 0;
}


static impulse_int impulse_transport_sock_join(impulse_int protocol,
					       impulse_char *node,
					       impulse_char *service,
					       impulse_unistring *address)
{
  impulse_int length = strlen(node) + strlen(service) + 10;
  impulse_char *buffer = impulse_alloc(length * sizeof(impulse_char));
  impulse_char *proto;
  
  proto = impulse_unistring_to_utf(sock_info[protocol].name, 8, 0);
  snprintf(buffer, length-1, "%s:%s:%s", proto, node, service);
  impulse_free(proto);
  *address = impulse_unistring_create_utf(buffer, 8);
  impulse_free(buffer);
  
  return 0;
}


static impulse_int impulse_transport_sock_query(impulse_int protocol,
						union sockaddr_all *addr,
						impulse_int addr_len,
						impulse_char **node,
						impulse_char **service)
{
  if (*node)
    impulse_free(*node);
  if (*service)
    impulse_free(*service);
  *node = NULL;
  *service = NULL;
  
  switch (addr->a.sa_family)
  {
#if HAVE_COMMS_SOCK_LOCAL
   case AF_LOCAL:
#elif HAVE_COMMS_SOCK_UNIX
   case AF_UNIX:
#endif
#if HAVE_COMMS_SOCK_LOCAL || HAVE_COMMS_SOCK_UNIX
    {
      char *node_buf = "local";
      *node = impulse_alloc(sizeof(impulse_char)*(strlen(node_buf)+1));
      strcpy(*node, node_buf);
      *service = impulse_alloc(sizeof(impulse_char)*(strlen(addr->un.sun_path)+1));
      strcpy(*service, addr->un.sun_path);
    }
    break;
#endif
    
#if HAVE_COMMS_SOCK_ADDRINFO
# if HAVE_COMMS_SOCK_IPV4
   case AF_INET:
# endif
# if HAVE_COMMS_SOCK_IPV6
   case AF_INET6:
# endif
# if HAVE_COMMS_SOCK_IPV4 || HAVE_COMMS_SOCK_IPV6
    {
      impulse_int ret = 0;
      impulse_char node_buf[NI_MAXHOST];
      impulse_char service_buf[NI_MAXSERV];
  
      if ((ret = getnameinfo(&addr->a, addr_len, 
			     node_buf, NI_MAXHOST, 
			     service_buf, NI_MAXSERV, 
			     (sock_info[protocol].type == SOCK_DGRAM) ? NI_DGRAM : 0)) < 0)
      {
	/* do something */
      }
      
      *node = impulse_alloc(strlen(node_buf)+1);
      *service = impulse_alloc(strlen(service_buf)+1);
      strcpy(*node, node_buf);
      strcpy(*service, service_buf);
    }
    break;
# endif
#else

# if HAVE_COMMS_SOCK_IPV4
   case AF_INET:
    {
      struct hostent *he = NULL;
      impulse_short port;
      impulse_char port_buf[50];
      
      if (!(he = gethostbyaddr((char *) &addr->in.sin_addr, sizeof(addr->in.sin_addr), AF_INET)))
	return -1;
      port = ntohs(addr->in.sin_port);
      sprintf(port_buf, "%d", port);
      *service = impulse_alloc(sizeof(impulse_char)*(strlen(port_buf)+1));
      strcpy(*service, port_buf);
      
      *node = impulse_alloc(sizeof(impulse_char)*(strlen(he->h_name)+1));
      strcpy(*node, he->h_name);
    }
    break;
# endif
    
# if HAVE_COMMS_SOCK_IPV6
   case AF_INET6:
    {
      struct hostent *he = NULL;
      impulse_short port;
      impulse_char port_buf[50];
      
      if (!(he = gethostbyaddr((char *) &addr->in6.sin6_addr, sizeof(addr->in6.sin6_addr), AF_INET6)))
	return -1;
      port = ntohs(addr->in6.sin6_port);
      sprintf(port_buf, "%d", port);
      *service = impulse_alloc(sizeof(impulse_char)*(strlen(port_buf)+1));
      strcpy(*service, port_buf);
      
      *node = impulse_alloc(sizeof(impulse_char)*(strlen(he->h_name)+1));
      strcpy(*node, he->h_name);
    }
    break;
# endif
#endif    
   default:
    {
      impulse_unistring desc = impulse_unistring_create_utf("Lookup of unknown address family requested", 8);
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
      impulse_unistring_destroy(desc);
    }
  }
  
  return 0;
}


static impulse_int impulse_transport_sock_lookup(impulse_int protocol,
						 impulse_char *node,
						 impulse_char *service,
						 union sockaddr_all **addr,
						 impulse_int *addr_len)
{
#if HAVE_COMMS_SOCK_ADDRINFO
  impulse_int ret = 0;
  struct addrinfo in;
  struct addrinfo *out;
  
  *addr = NULL;
  *addr_len = 0;

  in.ai_flags = 0;
  in.ai_socktype = sock_info[protocol].type;
  in.ai_family = sock_info[protocol].family;
  in.ai_protocol = sock_info[protocol].protocol;
  in.ai_addrlen = 0;
  in.ai_canonname = NULL;
  in.ai_addr = NULL;
  in.ai_next = NULL;
  
  if ((ret = getaddrinfo(node, service, &in, &out)) != 0)
    return -1;
  
  /*
   * Fix me: we should really walk through the linked
   * list and pick out the best entry.
   */
  *addr_len = out->ai_addrlen;
  *addr = impulse_alloc(*addr_len);
  memcpy(*addr, out->ai_addr, *addr_len);
  
  freeaddrinfo(out);
#else
  *addr = NULL;
  *addr_len = 0;
  
  if (impulse_transport_sock_split(address,
				   node,
				   service) < 0)
    return -1;
  
  if (!strcmp(*node, "local") ||
      !strcmp(*node, "unix"))
  {
    *addr = impulse_alloc(sizeof(union sockaddr_all));
# if HAVE_COMMS_SOCK_LOCAL
    (*addr)->un.sun_family = AF_LOCAL;
# elif HAVE_COMMS_SOCK_UNIX
    (*addr)->un.sun_family = AF_UNIX;
# else
    impulse_free(*addr);
    *addr = NULL;
    return -1;
# endif
    strncpy((*addr)->un.sun_path, *service, 107);
    (*addr)->un.sun_path[107] = '\0';
    *addr_len = SUN_LEN(&(*addr)->un);
  }
  else
  {
    struct hostent *he = NULL;
    struct servent *se = NULL;
    impulse_int port = 0;

    /*
     * Fix me: replace with reentrant ones if available
     * or make this stuff thread safe with mutexes
     */
    if (strlen(*node) && !(he = gethostbyname(*node)))
      return -1;
    if (strlen(*service))
    {
      impulse_char *end = NULL;
      port = strtoul(*service, &end, 10);
      if (end == *service)
      {
	if (!(se = getservbyname(*service, (type == SOCK_STREAM)?"tcp" : "udp")))
	  return -1;
	port = se->s_port;
      }
      else
	port = htons(port);
    }
    
    switch (he->h_addrtype)
    {
# if HAVE_COMMS_SOCK_IPV4
     case AF_INET:
      {
	(*addr) = impulse_alloc(sizeof(union sockaddr_all));
	*addr_len = sizeof(union sockaddr_all);
	
	(*addr)->in.sin_family = AF_INET;
	  (*addr)->in.sin_port = port;
	if (!strlen(*node))
	  (*addr)->in.sin_addr.s_addr = INADDR_ANY;
	else
	  (*addr)->in.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&((*addr)->in.sin_zero), 0, 8);
      }
      break;
# endif
    
# if HAVE_COMMS_SOCK_IPV6
     case AF_INET6:
      {
	(*addr) = impulse_alloc(sizeof(union sockaddr_all));
	*addr_len = sizeof(union sockaddr_all);

	(*addr)->in6.sin6_family = AF_INET6;
	(*addr)->in6.sin6_port = port;
	(*addr)->in6.sin6_flowinfo = 0;
	if (!strlen(*node))
	  (*addr)->in6.sin6_addr = in6addr_any;
	else
	  (*addr)->in6.sin6_addr = *((struct in6_addr *)he->h_addr);
	/* 
	 * This field is mentioned in RFC 2553, but does not seem to
	 * be implemented?
	 */
	//addr->in6.sin6_scope_id = 0;
      }
      break;
# endif
      
     default:
      return -1;
    }
  }
#endif
  return 0;
}


/*                            ======--------======                           */

void *impulse_transport_sock_create(void)
{
  impulse_transport_sock sock = impulse_alloc(sizeof(struct _impulse_transport_sock));
  
  sock->handle = -1;
  sock->protocol = -1;
  
  sock->node_local= NULL;
  sock->node_remote = NULL;
  sock->service_local = NULL;
  sock->service_remote = NULL;
  
  sock->handler = NULL;
  sock->edata = NULL;
  
  return sock;
}


void *impulse_transport_sock_create_handle(impulse_int handle,
					   impulse_int protocol)
{
  impulse_transport_sock sock = NULL;
  union sockaddr_all addr;
  impulse_int addr_len;
  
  sock = impulse_alloc(sizeof(struct _impulse_transport_sock));
  
  sock->handle = handle;
  sock->protocol = protocol;
  
  sock->node_local= NULL;
  sock->node_remote = NULL;
  sock->service_local = NULL;
  sock->service_remote = NULL;
  
  sock->handler = NULL;
  sock->edata = NULL;

  /* Fix me: don't return null */
  addr_len = sizeof(addr);
  if (getsockname(handle, &addr.a, &addr_len) < 0)
    return NULL;

  impulse_transport_sock_query(sock->protocol,
			       &addr,
			       addr_len,
			       &sock->node_local,
			       &sock->service_local);

  addr_len = sizeof(addr);
  if (getpeername(handle, &addr.a, &addr_len) < 0)
    return NULL;
  
  impulse_transport_sock_query(sock->protocol,
			       &addr,
			       addr_len,
			       &sock->node_remote,
			       &sock->service_remote);

  return sock;
}


void impulse_transport_sock_destroy(void *data)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  
  if (sock->handle != -1)
    impulse_transport_sock_close(data);
  
  impulse_free(sock);
}


static impulse_bool impulse_transport_sock_event(impulse_event *event,
						 void *data)
{
  if (event->type == IMPULSE_EVENT_FILEDESCRIPTOR &&
      event->data.fd.type == IMPULSE_EVENT_FD_READ)
  {
    impulse_transport_sock sock = (impulse_transport_sock)data;
    
    if (sock->handler)
      return (*sock->handler)(sock->edata);
  }
  return 0;
}


/* Event functions */
impulse_int impulse_transport_sock_handle(void *data,
					  impulse_transport_handler func,
					  void *edata)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  
  if (sock->handler)
    impulse_transport_sock_unhandle(data);

  sock->handler = func;
  sock->edata = edata;
  impulse_event_register_fd(sock->handle,
			    IMPULSE_EVENT_FD_READ,
			    impulse_transport_sock_event,
			    NULL,
			    data);
  
  return 0;
}


impulse_int impulse_transport_sock_unhandle(void *data)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  
  if (sock->handler)
  {
    impulse_event_deregister_fd(sock->handle);
    sock->handler = NULL;
    sock->edata = NULL;
  }
  
  return 0;
}



/* Connection functions */
impulse_int impulse_transport_sock_open(void *data,
					impulse_unistring local,
					impulse_unistring remote,
					impulse_int incoming)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  union sockaddr_all *addr_local = NULL;
  union sockaddr_all *addr_remote = NULL;
  impulse_int addr_len_local = 0;
  impulse_int addr_len_remote = 0;
  impulse_int reuseaddr = 1;
  
  if (sock->handle != -1)
    impulse_transport_sock_close(data);

  if (local)
  {
    if (impulse_transport_sock_split(local,
				     &sock->protocol,
				     &sock->node_local,
				     &sock->service_local) < 0)
      return -1;
    
    if (impulse_transport_sock_lookup(sock->protocol,
				      sock->node_local,
				      sock->service_local,
				      &addr_local,
				      &addr_len_local) < 0)
    {
      impulse_free(sock->node_local);
      impulse_free(sock->service_local);
      sock->node_local = NULL;
      sock->service_local = NULL;
      return -1;
    }
  }
  if (remote)
  {
    if (impulse_transport_sock_split(remote,
				     &sock->protocol,
				     &sock->node_remote,
				     &sock->service_remote) < 0)
    {
      impulse_free(addr_local);
      impulse_free(sock->node_local);
      impulse_free(sock->service_local);
      sock->node_local = NULL;
      sock->service_local = NULL;
      return -1;
    }
  
    if (impulse_transport_sock_lookup(sock->protocol,
				      sock->node_remote,
				      sock->service_remote,
				      &addr_remote,
				      &addr_len_remote) < 0)
    {
      impulse_free(addr_local);
      impulse_free(sock->node_local);
      impulse_free(sock->service_local);
      impulse_free(sock->node_remote);
      impulse_free(sock->service_remote);
      sock->node_local = NULL;
      sock->service_local = NULL;
      sock->node_remote = NULL;
      sock->service_remote = NULL;
      return -1;
    }
  }
  
  if ((sock->handle = socket(sock_info[sock->protocol].family,
			     sock_info[sock->protocol].type, 0)) == -1)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot create socket", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    /* Criticals never return */
  }

  if (setsockopt(sock->handle, SOL_SOCKET, SO_REUSEADDR, (void *)&reuseaddr, sizeof(reuseaddr)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot reuse socket address", 8);
    impulse_error_syscall(IMPULSE_ERROR_WARNING, desc);
    impulse_unistring_destroy(desc);
  }
  
  if (local)
  {
    if (bind(sock->handle, &addr_local->a, addr_len_local) < 0)
    {
      impulse_free(addr_local);
      impulse_free(addr_remote);
      
      close(sock->handle);
      sock->handle = -1;
      
      impulse_free(sock->node_local);
      impulse_free(sock->service_local);
      impulse_free(sock->node_remote);
      impulse_free(sock->service_remote);
      sock->node_local = NULL;
      sock->service_local = NULL;
      sock->node_remote = NULL;
      sock->service_remote = NULL;
      
      if (errno != EADDRINUSE)
      {
	impulse_unistring desc = impulse_unistring_create_utf("Cannot bind to server address", 8);
	impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
	impulse_unistring_destroy(desc);
      }
      return -1;
    }
  }
  

  if (incoming)
  {
    if (listen(sock->handle, incoming) < 0)
    {
      impulse_free(addr_local);
      impulse_free(addr_remote);
      
      close(sock->handle);
      sock->handle = -1;
      
      impulse_free(sock->node_local);
      impulse_free(sock->service_local);
      impulse_free(sock->node_remote);
      impulse_free(sock->service_remote);
      sock->node_local = NULL;
      sock->service_local = NULL;
      sock->node_remote = NULL;
      sock->service_remote = NULL;
      
      return -1;
    }
  }
  else
  {
    if (remote)
    {
      if (connect(sock->handle, &addr_remote->a, addr_len_remote) < 0)
      {
	impulse_free(addr_local);
	impulse_free(addr_remote);
	
	close(sock->handle);
	sock->handle = -1;
      
	impulse_free(sock->node_local);
	impulse_free(sock->service_local);
	impulse_free(sock->node_remote);
	impulse_free(sock->service_remote);
	sock->node_local = NULL;
	sock->service_local = NULL;
	sock->node_remote = NULL;
	sock->service_remote = NULL;
	
	return -1;
      }
    }
  }

  impulse_free(addr_local);
  impulse_free(addr_remote);
  
  return 0;
}


impulse_int impulse_transport_sock_accept(void *data,
					  void **client)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  impulse_int handle = -1;
  
  handle = accept(sock->handle, NULL, NULL);
  if (handle == -1)
    return -1;

  *client = impulse_transport_sock_create_handle(handle,
						 sock->protocol);
  
  return 0;
}


impulse_int impulse_transport_sock_close(void *data)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  
  if (sock->handle != -1)
  {
    close(sock->handle);
    sock->handle = -1;
    
    impulse_free(sock->node_local);
    impulse_free(sock->node_remote);
    impulse_free(sock->service_local);
    impulse_free(sock->service_remote);
    
    sock->node_local = NULL;
    sock->node_remote = NULL;
    sock->service_local = NULL;
    sock->service_remote = NULL;
  }
  return 0;
}


impulse_int impulse_transport_sock_active(void *data)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;

  return sock->handle != -1;
}


/* Data transfer functions */
impulse_int impulse_transport_sock_pending(void *data)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  impulse_int pending;
  impulse_int ret;
  
  ret = ioctl(sock->handle, FIONREAD, &pending);

  if (ret == -1)
    return -1;
  return pending;
}


impulse_int impulse_transport_sock_send(void *data,
					impulse_byte *buffer,
					impulse_uint size,
					impulse_unistring address)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  union sockaddr_all *addr = NULL;
  impulse_int addr_len = sizeof(addr);
  impulse_int ret = 0;
  impulse_int protocol;
  
  if (address)
  {
    impulse_char *node = NULL;
    impulse_char *service = NULL;

    impulse_transport_sock_split(address,
				 &protocol,
				 &node,
				 &service);
    impulse_transport_sock_lookup(sock->protocol,
				  node,
				  service,
				  &addr,
				  &addr_len);
    impulse_free(node);
    impulse_free(service);
  }
  
  if ((ret = sendto(sock->handle, buffer, size, 0, (address?&addr->a:NULL), (address?addr_len:0))) < 0)
    return -1;
  
  impulse_free(addr);
  return 0;
}


impulse_int impulse_transport_sock_receive(void *data,
					   impulse_byte *buffer,
					   impulse_uint size,
					   impulse_unistring *address)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  union sockaddr_all addr;
  impulse_int addr_len = sizeof(addr);
  impulse_int ret = 0;
  
  if ((ret = recvfrom(sock->handle, (char *)buffer, size, 0, &addr.a, &addr_len)) == -1)
    return -1;

  if (address)
  {
    impulse_char *node = NULL;
    impulse_char *service = NULL;
    
    impulse_transport_sock_query(sock->protocol,
				 &addr,
				 addr_len,
				 &node,
				 &service);
    impulse_transport_sock_join(sock->protocol,
				node,
				service,
				address);
    impulse_free(node);
    impulse_free(service);
  }

  return ret;
}


impulse_int impulse_transport_sock_peek(void *data,
					impulse_byte *buffer,
					impulse_uint size,
					impulse_unistring *address)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;
  union sockaddr_all addr;
  impulse_int addr_len = sizeof(addr);
  impulse_int ret;
  impulse_char *node = NULL;
  impulse_char *service = NULL;
  
  if ((ret = recvfrom(sock->handle, (char *)buffer, size, MSG_PEEK, &addr.a, &addr_len)) == -1)
    return -1;

  if (address)
  {
    impulse_transport_sock_query(sock->protocol,
				 &addr,
				 addr_len,
				 &node,
				 &service);
    impulse_transport_sock_join(sock->protocol,
				node,
				service,
				address);
  }

  return ret;
}


/* Address functions */
impulse_int impulse_transport_sock_local(void *data,
					 impulse_unistring *address)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;

  return impulse_transport_sock_join(sock->protocol,
				     sock->node_local,
				     sock->service_local,
				     address);
}


impulse_int impulse_transport_sock_remote(void *data,
					  impulse_unistring *address)
{
  impulse_transport_sock sock = (impulse_transport_sock)data;

  return impulse_transport_sock_join(sock->protocol,
				     sock->node_remote,
				     sock->service_remote,
				     address);
}

#else

impulse_transport impulse_transport_sock = NULL;

#endif

/* End of file */
