/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * child_xwin.c
 * ------------
 *
 * Architecture-specific child functions
 */

#include "impulse/util/event.h"
#include "impulse/util/memory.h"
#include "impulse/rows/child_xwin.h"
#include "impulse/rows/main_xwin.h"
#include "impulse/rows/screen.h"

static impulse_bool _p_child_expose(XEvent     *ev,
				    ROWS_child  child);
static impulse_bool _p_child_configure(XEvent     *ev,
				       ROWS_child  child);
static impulse_bool _p_child_visibility(XEvent     *ev,
					ROWS_child  child);
static impulse_bool _p_child_crossing(XEvent     *ev,
				      ROWS_child  child);
static impulse_bool _p_child_button(XEvent     *ev,
				    ROWS_child  child);
static impulse_bool _p_child_motion(XEvent     *ev,
				    ROWS_child  child);

static struct _p_ROWS_X_child_events _p_child_events = 
{
  {
    NULL, /* No event */
    NULL, /* Error event */
    NULL, /* Key press */
    NULL, /* Key release */
    _p_child_button, /* Button press */
    _p_child_button, /* Button release */
    _p_child_motion, /* Motion notify */
    _p_child_crossing, /* Enter notify */
    _p_child_crossing, /* Leave notify */
    NULL, /* Focus in */
    NULL, /* Focus out */
    NULL, /* Keymap notify */
    _p_child_expose, /* Expose notify */
    NULL, /* Graphics expose */
    NULL, /* No expose */
    NULL, /* Visibility notify */
    NULL, /* Create notify */
    NULL, /* Destroy notify */
    _p_child_visibility, /* Unmap notify */
    _p_child_visibility, /* Map notify */
    NULL, /* Map request */
    NULL, /* Reparent notify */
    _p_child_configure, /* Configure notify */
    NULL, /* Configure request */
    NULL, /* Gravity notify */
    NULL, /* Resize request */
    NULL, /* Circulate notify */
    NULL, /* Circulate request */
    NULL, /* Property notify */
    NULL, /* Selection clear */
    NULL, /* Selection request */
    NULL, /* Selection notify */
    NULL, /* Colourmap notify */
    NULL, /* Client message */
    NULL, /* Mapping notify */
  }
};

static impulse_int _p_child_mask = ButtonPressMask | ButtonReleaseMask |
PointerMotionMask | ExposureMask | EnterWindowMask |
LeaveWindowMask | StructureNotifyMask;



/*                            ======--------======                           */

static impulse_bool window_event(impulse_event *event,
				void *data);
static void ROWS_child_int_actual_widget(ROWS_child child);

/*                            ======--------======                           */

void _p_ROWS_child_int_setup(ROWS_widget   *widget,
			     ROWS_drawable  drawable)
{
  _p_ROWS_child_data   **private;
  _p_ROWS_X_child_data  *data;

  private = (_p_ROWS_child_data **) ROWS_widget_private(widget,
							_p_ROWS_child_ID);

  data = (*private)->int_data =
    impulse_alloc(sizeof(struct _p_ROWS_X_child_data));

  data->window     = None;
  data->parent     = None;
  data->display    = appl->display;
  data->nevents = 0;
  data->events = NULL;
  data->mask = 0;
}


void _p_ROWS_child_int_destroy(_p_ROWS_child_data **data)
{
  _p_ROWS_X_child_data  *x_data;

  x_data = (*data)->int_data;
  if (x_data->window != None)
    XDestroyWindow(x_data->display,
		   x_data->window);
  
  impulse_free((*data)->int_data);
}


void _p_ROWS_child_int_container(ROWS_child     child,
				 ROWS_container container)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *data;
 
  Window container_win;

  priv = (_p_ROWS_child_data **) child;

  data = (*priv)->int_data;

  data->parent =
    container_win = ROWS_X_child_get_window(ROWS_child_cast(container));

  if ((*priv)->actualised)
    ROWS_child_int_actual_widget(child);
  else
    data->window = data->parent;
}

/*                            ======--------======                           */

Window ROWS_X_child_get_window(ROWS_child child)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;
  Window                 win;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;
  
  x_data = (*priv)->int_data;

  win = x_data->window;

  ROWS_child_unlock(child);

  return win;
}

void ROWS_X_child_set_window(ROWS_child child,
			     Window     win)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;
  
  x_data = (*priv)->int_data;

  if (x_data->window == x_data->parent)
    x_data->parent = win;

  x_data->window = win;

  (*priv)->draw = ROWS_drawable_cast(ROWS_screen_create(child));
  (*priv)->style = ROWS_drawable_get_style((*priv)->draw);
  
  ROWS_child_unlock(child);
}


extern void ROWS_X_child_register_events(ROWS_child child,
					 _p_ROWS_X_child_events events,
					 impulse_int mask)
{
  _p_ROWS_child_data **priv;
  _p_ROWS_X_child_data  *x_data;
  
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;
  
  x_data = (*priv)->int_data;

  x_data->mask |= mask;
  x_data->nevents++;
  if (x_data->events)
    x_data->events = impulse_realloc(x_data->events,
				     sizeof(_p_ROWS_X_child_events) * x_data->nevents);
  else
    x_data->events = impulse_alloc(sizeof(_p_ROWS_X_child_events) * x_data->nevents);
  x_data->events[x_data->nevents-1] = events;

  if (x_data->window)
    XSelectInput(x_data->display,
		 x_data->window,
		 x_data->mask);

  ROWS_child_unlock(child);
}


static impulse_bool _p_ROWS_X_child_move_window(ROWS_child child,
						ROWS_child_event_num type,
						ROWS_child_event *event,
						void *data)
{
  _p_ROWS_child_data **priv;
  _p_ROWS_X_child_data  *x_data;
  ROWS_point     *origin;
  XWindowChanges  changes;
  
  priv = (_p_ROWS_child_data **) child;  
  x_data = (*priv)->int_data;

  origin = ROWS_box_get_position(ROWS_box_cast(child));

  changes.x      = MICRONS_TO_PIXEL(origin->x);
  changes.y      = MICRONS_TO_PIXEL(origin->y);

  XConfigureWindow(x_data->display,
		   x_data->window,
		   CWX|CWY,
		   &changes);

  XFlush(x_data->display);
  return 0;
}


static impulse_bool _p_ROWS_X_child_size_window(ROWS_child child,
						ROWS_child_event_num type,
						ROWS_child_event *event,
						void *data)
{
  _p_ROWS_child_data **priv;
  _p_ROWS_X_child_data  *x_data;
  ROWS_point     *size;
  XWindowChanges  changes;
  
  priv = (_p_ROWS_child_data **) child;  
  x_data = (*priv)->int_data;

  size   = ROWS_box_get_physical(ROWS_box_cast(child));
  
  if (size->x <= 0)
    size->x = PIXEL_TO_MICRONS(2);
  if (size->y <= 0)
    size->y = PIXEL_TO_MICRONS(2);
  
  changes.width  = MICRONS_TO_PIXEL(size->x);
  changes.height = MICRONS_TO_PIXEL(size->y);
  
  XConfigureWindow(x_data->display,
		   x_data->window,
		   CWWidth|CWHeight,
		   &changes);

  XFlush(x_data->display);
  return 0;
}


void ROWS_X_child_window_attributes(ROWS_child child)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;
  XSetWindowAttributes   attributes;
  ROWS_style            style;
 
  ROWS_child_lock(child);
  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  style = ROWS_child_get_style(child);

  priv = (_p_ROWS_child_data **) child;  
  x_data = (*priv)->int_data;

  attributes.colormap         = appl->cmap;
  attributes.background_pixel = ROWS_brush_get_fgpixel(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_BACKGROUND]);
  attributes.border_pixel     = ROWS_brush_get_fgpixel(style->brush[ROWS_STYLE_NORMAL][ROWS_STYLE_BACKGROUND]);

  XChangeWindowAttributes(x_data->display,
			  x_data->window,
			  CWBackPixel | CWBorderPixel | CWColormap,
			  &attributes);

  XFlush(x_data->display);

  ROWS_child_unlock(child);
}


void ROWS_X_child_create_window(ROWS_child child,
				Window     parent)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;
  Window                 win;
  ROWS_point     *origin;
  ROWS_point     *size;

  ROWS_child_lock(child);

  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;
  x_data = (*priv)->int_data;

  if (!x_data->parent)
    x_data->parent = DefaultRootWindow(x_data->display);

  origin = ROWS_box_get_position(ROWS_box_cast(child));
  size   = ROWS_box_get_physical(ROWS_box_cast(child));
  
  if (size->x <= 0)
    size->x = PIXEL_TO_MICRONS(2);
  if (size->y <= 0)
    size->y = PIXEL_TO_MICRONS(2);
  
  x_data->window =
    win = XCreateSimpleWindow(x_data->display,
			      x_data->parent,
			      MICRONS_TO_PIXEL(origin->x),
			      MICRONS_TO_PIXEL(origin->y),
			      MICRONS_TO_PIXEL(size->x),
			      MICRONS_TO_PIXEL(size->y),
			      0,
			      BlackPixel(x_data->display, 0),
			      WhitePixel(x_data->display, 0));

  impulse_event_register(IMPULSE_EVENT_GUI,
			 win,
			 window_event,
			 child);

  ROWS_X_child_register_events(child,
			       &_p_child_events,
			       _p_child_mask);

  if (x_data->parent == DefaultRootWindow(x_data->display))
  {
    ROWS_child_event_register(child,
			      ROWS_CHILD_PREMOVE,
			      _p_ROWS_X_child_move_window,
			      NULL);
    ROWS_child_event_register(child,
			      ROWS_CHILD_PRESIZE,
			      _p_ROWS_X_child_size_window,
			      NULL);
  }
  else
  {
    ROWS_child_event_register(child,
			      ROWS_CHILD_MOVE,
			      _p_ROWS_X_child_move_window,
			      NULL);
    ROWS_child_event_register(child,
			      ROWS_CHILD_SIZE,
			      _p_ROWS_X_child_size_window,
			      NULL);
  }

  (*priv)->offset.x = 0;
  (*priv)->offset.y = 0;
  (*priv)->draw = ROWS_drawable_cast(ROWS_screen_create(child));
  (*priv)->style = ROWS_drawable_get_style((*priv)->draw);
  
  ROWS_X_child_window_attributes(child);

  ROWS_child_unlock(child);
}

void ROWS_child_int_actual_widget(ROWS_child child)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;
  Window                 win;
  
  ROWS_child_lock(child);

  ROWS_widget_check_type(child, _p_ROWS_child_type, "ROWS_child");

  priv = (_p_ROWS_child_data **) child;  
  x_data = (*priv)->int_data;

  if (x_data->window == None || x_data->window == x_data->parent)
    {
      ROWS_X_child_create_window(child,
				 x_data->parent);
      win = x_data->window;
      (*priv)->offset.x = 0;
      (*priv)->offset.y = 0;
      (*priv)->draw = ROWS_drawable_cast(ROWS_screen_create(child));
      (*priv)->style = ROWS_drawable_get_style((*priv)->draw);

      if ((*priv)->visible)
	XMapWindow(x_data->display, win);
    }
  else
    {
      win = x_data->window;
      XUnmapWindow(x_data->display, win);
      XReparentWindow(x_data->display,
		      win,
		      x_data->parent,
		      0,0);

      if ((*priv)->visible)
	XMapWindow(x_data->display, win);
    }

  ROWS_child_unlock(child);
}


void _p_ROWS_child_int_update(ROWS_child child)
{
  _p_ROWS_child_data     **priv;
  _p_ROWS_X_child_data    *x_data;
  ROWS_style_active_type   type;
  ROWS_style              style;
  XSetWindowAttributes     attributes;  

  priv = (_p_ROWS_child_data **) child;  
  x_data  = (*priv)->int_data;

  type    = ROWS_child_get_activation(child);
  style   = ROWS_child_get_style(child);

  if ((*priv)->actualised)
    {
      if (x_data->window != None && x_data->window != x_data->parent)
	{
	  attributes.background_pixel =
	    ROWS_brush_get_fgpixel(style->brush[type][ROWS_STYLE_BACKGROUND]);
	  attributes.border_pixel     =
	    ROWS_brush_get_fgpixel(style->brush[type][ROWS_STYLE_BACKGROUND]);

	  XChangeWindowAttributes(x_data->display,
				  x_data->window,
				  CWBackPixel | CWBorderPixel,
				  &attributes);
//	  XClearArea(x_data->display,
//		     x_data->window,
//		     0,0,0,0,
//		     True);
	}
    }
}

void _p_ROWS_child_int_set_visible(ROWS_child child,
				   impulse_bool visible)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;

  priv = (_p_ROWS_child_data **) child;  
  x_data = (*priv)->int_data;

  if ((*priv)->actualised)
  {
    if (x_data->window != None && x_data->window != x_data->parent)
    {
      if (visible)
	XMapWindow(x_data->display, x_data->window);
      else
	XUnmapWindow(x_data->display, x_data->window);
    }
    /*
     * The visibility event will be triggered asynchronously
     * when X gets round to actually mapping the window.
     * We need this, cos X will go a bit mad if we try to
     * draw on an unmapped window.
     */
  }
  else
  {
    ROWS_child_event event;
    (*priv)->visible = visible;
    event.visibility.visible = visible;
    ROWS_child_event_send(child, ROWS_CHILD_VISIBILITY, &event);
  }
}


void _p_ROWS_child_int_set_realised(ROWS_child child,
				    impulse_bool realised)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;
  ROWS_child_event event;

  priv = (_p_ROWS_child_data **) child;  
  x_data = (*priv)->int_data;

  if ((*priv)->actualised)
    ROWS_child_int_actual_widget(child);

  (*priv)->realised = realised;
  event.realisation.realised = realised;
  ROWS_child_event_send(child, ROWS_CHILD_REALISATION, &event);
}


void _p_ROWS_child_int_set_actualised(ROWS_child child,
				      impulse_bool actualised)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data  *x_data;

  priv = (_p_ROWS_child_data **) child;  
  x_data = (*priv)->int_data;

  (*priv)->actualised = actualised;
  if ((*priv)->realised)
    ROWS_child_int_actual_widget(child);
}


/*                            ======--------======                           */

static impulse_bool _p_child_expose_pred(impulse_event *event,
					 void          *data)
{
  _p_ROWS_X_child_data *x_data;

  if (event->type != IMPULSE_EVENT_GUI)
    return 0;
  
  x_data = data;

  if (event->data.gui.object == x_data->window &&
      event->data.gui.type == Expose)
      return 1;
  else
    return 0;
}

static impulse_bool _p_child_expose(XEvent     *ev,
				    ROWS_child  child)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data   *x_data;
  ROWS_child_event        event; 
  ROWS_drawable           drawable;
  XEvent *free_me = NULL;
  
  priv = (_p_ROWS_child_data **) child;
  x_data  = (*priv)->int_data;

  drawable = ROWS_child_get_drawable(ROWS_child_cast(child));
  
  ROWS_drawable_clip_push(drawable);
  ROWS_drawable_clip_clear(drawable);

  for (;;)
  {
    ROWS_rect rect;
    impulse_event next;
    
    rect.pos.x  = PIXEL_TO_MICRONS(ev->xexpose.x);
    rect.pos.y  = PIXEL_TO_MICRONS(ev->xexpose.y);
    rect.size.x = PIXEL_TO_MICRONS(ev->xexpose.width);
    rect.size.y = PIXEL_TO_MICRONS(ev->xexpose.height);
    
    ROWS_drawable_clip_union(drawable, &rect);

    if (!ev->xexpose.count ||
	!impulse_event_check_next(&next,
				 _p_child_expose_pred,
				 priv))
      break;
    
    if (free_me)
      impulse_free(free_me);
    ev = next.data.gui.data;
    free_me = ev;
  }
  
  if (free_me)
    impulse_free(free_me);

  ROWS_drawable_clip_set(drawable);
  ROWS_drawable_clip_bounds(drawable, &event.exposure.clip);

  ROWS_child_event_send(child,
			ROWS_CHILD_EXPOSURE,
			&event);

  XFlush(x_data->display);
  ROWS_drawable_clip_pop(drawable);

  return 1;
}


static impulse_bool _p_child_configure(XEvent  *event,
				       ROWS_child child)
{
  _p_ROWS_child_data   **priv;
  _p_ROWS_X_child_data   *x_data;
  ROWS_box         box;
  ROWS_child_event ev;

  priv = (_p_ROWS_child_data **) child;
  x_data  = (*priv)->int_data;

  if (x_data->parent != DefaultRootWindow(x_data->display))
    return 0;

  box = ROWS_box_cast(child);

  ev.size.size.x = PIXEL_TO_MICRONS(event->xconfigure.width);
  ev.size.size.y = PIXEL_TO_MICRONS(event->xconfigure.height);

  ROWS_box_set_physical(box, &ev.size.size);
  ROWS_box_set_virtual (box, &ev.size.size);
  
  ROWS_child_event_send(child,
			ROWS_CHILD_SIZE,
			&ev);

  ev.move.position.x = PIXEL_TO_MICRONS(event->xconfigure.x);
  ev.move.position.y = PIXEL_TO_MICRONS(event->xconfigure.y);

  ROWS_box_set_position(box, &ev.move.position);
  
  ROWS_child_event_send(child,
			ROWS_CHILD_MOVE,
			&ev);

  return 1;
}


static impulse_bool _p_child_visibility(XEvent     *ev,
					ROWS_child  child)
{
  ROWS_child_event event;
  _p_ROWS_child_data **priv = (_p_ROWS_child_data **)child;
  
  (*priv)->visible = (ev->type == MapNotify) ? 1 : 0;
  event.visibility.visible = (ev->type == MapNotify) ? 1 : 0;
  ROWS_child_event_send(child,
			ROWS_CHILD_VISIBILITY,
			&event);
  return 1;
}


static impulse_bool _p_child_crossing(XEvent     *ev,
				      ROWS_child  child)
{
  ROWS_child_event event;
  
  event.crossing.entered = (ev->type == EnterNotify) ? 1 : 0;
  event.crossing.position.x = PIXEL_TO_MICRONS(ev->xcrossing.x);
  event.crossing.position.y = PIXEL_TO_MICRONS(ev->xcrossing.y);
  event.crossing.state = ROWS_X_get_state(ev->xcrossing.state);
  
  ROWS_child_event_send(child,
			ROWS_CHILD_CROSSING,
			&event);
  return 1;
}


static impulse_bool _p_child_button(XEvent     *ev,
				    ROWS_child  child)
{
  ROWS_child_event event;
  
  event.button.pressed = (ev->type == ButtonPress) ? 1 : 0;
  event.button.position.x = PIXEL_TO_MICRONS(ev->xbutton.x);
  event.button.position.y = PIXEL_TO_MICRONS(ev->xbutton.y);
  event.button.button = ROWS_X_get_button(ev->xbutton.button);
  
  /*
   * Fudge it so state includes the button that was actually pressed/released
   */
  if (ev->type == ButtonPress)
    event.button.state = ROWS_X_get_state(ev->xbutton.state) | (1 << (event.button.button-1));
  else
    event.button.state = ROWS_X_get_state(ev->xbutton.state) & ~(1 << (event.button.button -1));
  
  ROWS_child_event_send(child,
			ROWS_CHILD_BUTTON,
			&event);
  return 1;
}


static impulse_bool _p_child_motion(XEvent     *ev,
				    ROWS_child  child)
{
  ROWS_child_event event;
  
  event.motion.position.x = PIXEL_TO_MICRONS(ev->xmotion.x);
  event.motion.position.y = PIXEL_TO_MICRONS(ev->xmotion.y);
  event.motion.state = ROWS_X_get_state(ev->xmotion.state);
  
  ROWS_child_event_send(child,
			ROWS_CHILD_MOTION,
			&event);
  return 1;
}


static impulse_bool window_event(impulse_event  *event,
				void *data)
{
  XEvent              *ev;
  ROWS_child child = data;
  _p_ROWS_child_data **priv;
  _p_ROWS_X_child_data *x_data;
  impulse_int i;
  
  ev = event->data.gui.data;

  ROWS_child_lock(child);

  priv = (_p_ROWS_child_data **)child;
  x_data = (*priv)->int_data;

//  printf("**** Got event %x %d\n", ev->xany.window, ev->xany.type);
  
  for (i = 0 ; i < x_data->nevents ; i++)
  {
    if (x_data->events[i]->funcs[ev->type])
      if ((*x_data->events[i]->funcs[ev->type])(ev, child))
	break;
  }
  
  ROWS_child_unlock(child);
  impulse_free(ev);

  return 1;
}
