/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * eventbox.c
 * ----------
 *
 * A box to which events may occur (needs a seperate event delivery system,
 * though)
 */

#include "../../../config.h"

#include "impulse/rows/eventbox_int.h"
#include "impulse/util/memory.h"

/*                            ======--------======                           */

static void ROWS_eventbox_setup(ROWS_widget *widget,
				int          widget_level);
static int  eventbox_ID   = -1;
static int  eventbox_type = -1;

static void ROWS_eventbox_destroy(void **priv);

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget ROWS_eventbox_create(void)
{
  ROWS_widget *wid;

  if (eventbox_ID   == -1)
    eventbox_ID   = ROWS_box_next_identifier();

  wid = ROWS_eventbox_initialise(eventbox_ID);

  return *wid;
}


ROWS_widget *ROWS_eventbox_initialise(int id)
{
  ROWS_widget *widget;

  if (eventbox_ID   == -1)
    eventbox_ID   = ROWS_box_next_identifier();
  if (eventbox_type == -1)
    eventbox_type = ROWS_widget_next_type();

  widget = ROWS_box_initialise(id);
  ROWS_widget_set_type(widget, eventbox_ID, eventbox_type);
  ROWS_widget_set_destructor(widget, eventbox_ID, ROWS_eventbox_destroy);

  ROWS_eventbox_setup(widget, id);

  return widget;
}


int ROWS_eventbox_next_identifier(void)
{
  if (eventbox_ID==-1)
    eventbox_ID = ROWS_box_next_identifier();

  return eventbox_ID + 1;
}


ROWS_eventbox ROWS_eventbox_cast(void **widget)
{
  if (eventbox_ID   == -1)
    eventbox_ID   = ROWS_box_next_identifier();
  if (eventbox_type == -1)
    eventbox_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, eventbox_ID, eventbox_type);
}


void ROWS_eventbox_setup(ROWS_widget *widget,
			 int          widget_level)
{
  _p_ROWS_eventbox_data **priv;
  int x;

  priv  = (_p_ROWS_eventbox_data **) ROWS_widget_private(widget, eventbox_ID);
  *priv = impulse_alloc(sizeof(_p_ROWS_eventbox_data));

  (*priv)->events =
    impulse_alloc(sizeof(_p_ROWS_eventbox_array ***)*(widget_level+1));
  (*priv)->info   = impulse_alloc(sizeof(_p_ROWS_eventbox_widget_info)*(widget_level+1));

  (*priv)->widget_level = widget_level;

  for (x=0;x<=widget_level;x++)
    {
      (*priv)->events[x]        = NULL;
      (*priv)->info[x].maxevent = -1;
    }
}


void ROWS_eventbox_destroy(void **data)
{
  int x,y;
  _p_ROWS_eventbox_array  *next, *ths;
  _p_ROWS_eventbox_data  **priv;

  priv = (_p_ROWS_eventbox_data **) data;

  for (x=0;x<=(*priv)->widget_level;x++)
    {
      for (y=0;y<=(*priv)->info[x].maxevent;y++)
	{
	  ths = (*priv)->events[x][y];
	  while (ths!=NULL)
	    {
	      next = ths->next;

	      impulse_free(ths);
	      
	      ths = next;
	    }
	}

      impulse_free((*priv)->events[x]);
    }

  impulse_free((*priv)->info);
  impulse_free((*priv)->events);
  impulse_free(*priv);
}

/*                            ======--------======                           */

/* Class-specific functions */

void ROWS_eventbox_register_widget(ROWS_eventbox box,
				   int widget_level,
				   int max_event)
{
  _p_ROWS_eventbox_data **priv;
  int                     x;

  ROWS_widget_lock(box);
  
  ROWS_widget_check_type(box, eventbox_type, "ROWS_eventbox");
  priv = (_p_ROWS_eventbox_data **) box;

  (*priv)->events[widget_level] =
    impulse_alloc(sizeof(_p_ROWS_eventbox_array **)*(max_event+1));
  (*priv)->info[widget_level].maxevent = max_event;

  for (x=0;x<(max_event+1);x++)
    (*priv)->events[widget_level][x] = NULL;
  
  ROWS_widget_unlock(box);
}


void ROWS_eventbox_register_handler(ROWS_eventbox box,
				    int widget_level,
				    ROWS_eventbox_func function,
				    int  event_number,
				    void *data)
{
  _p_ROWS_eventbox_data **priv;
  _p_ROWS_eventbox_array *next, *temp;
  
  ROWS_widget_lock(box);

#if 1
  ROWS_widget_check_type(box, eventbox_type, "ROWS_eventbox");
  priv = (_p_ROWS_eventbox_data **) box;

  temp = impulse_alloc(sizeof(_p_ROWS_eventbox_array));
  temp->next = (*priv)->events[widget_level][event_number];
  temp->func = function;
  temp->data = data;

  (*priv)->events[widget_level][event_number] = temp;
#else
  next = (*priv)->events[widget_level][event_number];
  while (next && next->next)
    next = next->next;
  
  temp = impulse_alloc(sizeof(_p_ROWS_eventbox_array));
  temp->next = NULL;
  temp->func = function;
  temp->data = data;
  
  if (next)
    next->next = temp;
  else
    (*priv)->events[widget_level][event_number] = temp;
#endif
  
  ROWS_widget_unlock(box);
}


void ROWS_eventbox_deregister_handler(ROWS_eventbox box,
				      int widget_level,
				      ROWS_eventbox_func function,
				      int  event_number,
				      void *data)
{
  _p_ROWS_eventbox_data **priv;
  _p_ROWS_eventbox_array *next, *prev;
  
  ROWS_widget_lock(box);

  ROWS_widget_check_type(box, eventbox_type, "ROWS_eventbox");
  priv = (_p_ROWS_eventbox_data **) box;

  next = (*priv)->events[widget_level][event_number];
  prev = NULL;
  while (next)
  {
    if (next->func == function &&
	next->data == data)
    {
      if (prev)
	prev->next = next->next;
      else
	(*priv)->events[widget_level][event_number] = next->next;
      impulse_free(next);
      break;
    }
  }
  
  ROWS_widget_unlock(box);
}


/* Note: Assumes widget is already locked */
void ROWS_eventbox_send(ROWS_eventbox box,
			int    widget_level,
			void **widget,
			int    event_number,
			void  *event_data)
{
  _p_ROWS_eventbox_data **priv;
  _p_ROWS_eventbox_array *next;
  int done;
  
  /* ROWS_widget_lock(box); */

  /* 
   * Want to know what I was on about in ROWSchild.c? (ROWS_child_lock)
   * Uncomment the usleep statement - it aggravates the problem.
   */

  /*
  usleep(1);
  */

  ROWS_widget_check_type(box, eventbox_type, "ROWS_eventbox");
  priv = (_p_ROWS_eventbox_data **) box;

  next = (*priv)->events[widget_level][event_number];
  done = 0;
  
  while (next!=NULL && !done)
    {
      done = (next->func)(widget, event_number, event_data, next->data);
      next = next->next;
    }
   
  /* ROWS_widget_unlock(box); */
}


impulse_bool ROWS_eventbox_is_handled(ROWS_eventbox box,
				      int widget_level,
				      int event_number)
{
  _p_ROWS_eventbox_data **priv;
  impulse_bool found = 0;
  
  ROWS_widget_lock(box);
  
  ROWS_widget_check_type(box, eventbox_type, "ROWS_eventbox");
  priv = (_p_ROWS_eventbox_data **) box;

  found = ((*priv)->events[widget_level][event_number] != NULL);
  
  ROWS_widget_unlock(box);
  
  return found;
}


/* End of file */
