/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * window_xwin.c
 * -------------
 *
 * Implementation of windows for X
 */

#include "impulse/rows/window_xwin.h"
#include "impulse/rows/main_xwin.h"
#include "impulse/rows/child_xwin.h"
#include "impulse/rows/child_int.h"
#include "impulse/util/event.h"

/*                            ======--------======                           */

extern int _p_ROWS_window_type;

/*                            ======--------======                           */

static impulse_bool _p_window_visibility(XEvent  *event,
					 ROWS_widget widget);
static impulse_bool  _p_window_key(XEvent  *event,
				   ROWS_widget widget);
static impulse_bool  _p_window_client(XEvent  *event,
				      ROWS_widget widget);
static impulse_bool  _p_window_focus(XEvent  *event,
				     ROWS_widget widget);


static struct _p_ROWS_X_child_events _p_window_events =
{
  {
    NULL, /* No event */
    NULL, /* Error event */
    _p_window_key, /* Key press */
    _p_window_key, /* Key release */
    NULL, /* Button press */
    NULL, /* Button release */
    NULL, /* Motion notify */
    NULL, /* Enter notify */
    NULL, /* Leave notify */
    _p_window_focus, /* Focus in */
    _p_window_focus, /* Focus out */
    NULL, /* Keymap notify */
    NULL, /* Expose notify */
    NULL, /* Graphics expose */
    NULL, /* No expose */
    NULL, /* Visibility notify */
    NULL, /* Create notify */
    NULL, /* Destroy notify */
    _p_window_visibility, /* Unmap notify */
    _p_window_visibility, /* Map notify */
    NULL, /* Map request */
    NULL, /* Reparent notify */
    NULL, /* Configure notify */
    NULL, /* Configure request */
    NULL, /* Gravity notify */
    NULL, /* Resize request */
    NULL, /* Circulate notify */
    NULL, /* Circulate request */
    NULL, /* Property notify */
    NULL, /* Selection clear */
    NULL, /* Selection request */
    NULL, /* Selection notify */
    NULL, /* Colourmap notify */
    _p_window_client, /* Client message */
    NULL, /* Mapping notify */
  }
};

static impulse_int _p_window_mask = KeyPressMask | KeyReleaseMask |
FocusChangeMask | StructureNotifyMask;



/*                            ======--------======                           */

/* Class-specific functions */

void ROWS_window_set_title(ROWS_window  window,
			   char        *title)
{
  _p_ROWS_window_data **private;

  ROWS_widget_check_type(window, _p_ROWS_window_type, "ROWS_window");

  ROWS_widget_lock(window);
  private = (_p_ROWS_window_data **) window;

  impulse_free((*private)->properties.title);
  if (title==NULL)
    title = "";

  (*private)->properties.title = impulse_alloc(strlen(title)+1);
  strcpy((*private)->properties.title, title);

  ROWS_widget_unlock(window);
}

void ROWS_window_set_class(ROWS_window  window,
			   char        *class)
{
  _p_ROWS_window_data **private;

  ROWS_widget_check_type(window, _p_ROWS_window_type, "ROWS_window");

  ROWS_widget_lock(window);
  private = (_p_ROWS_window_data **) window;

  impulse_free((*private)->properties.class);
  if (class==NULL)
    class = "";

  (*private)->properties.class = impulse_alloc(strlen(class)+1);
  strcpy((*private)->properties.class, class);

  ROWS_widget_unlock(window);
}

void ROWS_window_set_icon_name(ROWS_window  window,
			       char        *icon_name)
{
  _p_ROWS_window_data **private;

  ROWS_widget_check_type(window, _p_ROWS_window_type, "ROWS_window");

  ROWS_widget_lock(window);
  private = (_p_ROWS_window_data **) window;

  impulse_free((*private)->properties.icon_name);
  if (icon_name==NULL)
    icon_name = "";

  (*private)->properties.icon_name = impulse_alloc(strlen(icon_name)+1);
  strcpy((*private)->properties.icon_name, icon_name);

  ROWS_widget_unlock(window);
}

/*                            ======--------======                           */

/* Event functions */

void ROWS_window_int_setup(ROWS_widget          *widget,
			   _p_ROWS_window_data  *data)
{
  ROWS_point size;
  ROWS_child_set_actualised(ROWS_child_cast(*widget),1);
  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_REALISATION,
			    _p_ROWS_window_realise,
			    NULL);
  size.x = PIXEL_TO_MICRONS(200);
  size.y = PIXEL_TO_MICRONS(200);
  ROWS_box_set_physical(ROWS_box_cast(*widget),
			&size);
}


void ROWS_window_int_destroy(_p_ROWS_window_data **private)
{
}


/*                            ======--------======                           */

impulse_bool  _p_ROWS_window_realise(ROWS_child            widget,
				     ROWS_child_event_num  event_number,
				     ROWS_child_event     *event,
				     void                 *data)
{
  _p_ROWS_window_data **priv;
  _p_ROWS_child_data  **cpriv;
  _p_ROWS_X_child_data *x_data;
  Atom                   protocols[2];
  XTextProperty          textproperty;
  XClassHint            *classhint;

  priv = (_p_ROWS_window_data **)ROWS_window_cast(widget);
  cpriv = (_p_ROWS_child_data **)widget;
  x_data = (*cpriv)->int_data;
  
  if (!event->realisation.realised)
    return 0;
  
  /* Set our window up properly */
  ROWS_X_child_register_events(ROWS_child_cast(widget),
			       &_p_window_events,
			       _p_window_mask);

  impulse_mutex_X_lock;
  /* Set window properties */

  /* Titles, icon names, etc */
  XStringListToTextProperty(&(*priv)->properties.title, 1, &textproperty);
  XSetWMName(x_data->display,
	     x_data->window,
	     &textproperty);
  XFree(textproperty.value);

  XStringListToTextProperty(&(*priv)->properties.icon_name,
			    1, &textproperty);
  XSetWMIconName(x_data->display,
		 x_data->window,
		 &textproperty);
  XFree(textproperty.value);

  /* FIXME: Check for NULL values */
  classhint = XAllocClassHint();
  classhint->res_class = impulse_alloc(strlen((*priv)->properties.class)+1);
  classhint->res_name = "ROWS Application";
  strcpy(classhint->res_class, (*priv)->properties.class);
  XSetClassHint(x_data->display, x_data->window, classhint);
  impulse_free(classhint->res_class);
  XFree(classhint);

  /* Protocols */
  protocols[0] = XInternAtom(x_data->display,
			     "WM_DELETE_WINDOW",
			     1);
  /* protocols[1] = XInternAtom((*private)->display,
			     "WM_TAKE_FOCUS",
			     1); */
  XSetWMProtocols(x_data->display,
		  x_data->window,
		  protocols,
		  1);
  impulse_mutex_X_unlock;

  return 0;
}


/*                            ======--------======                           */

/* X Event functions */

static impulse_bool  _p_window_visibility(XEvent  *event,
					  ROWS_widget widget)
{
  ROWS_drawable_set_valid(ROWS_child_get_drawable(ROWS_child_cast(widget)), 
			  (event->type == MapNotify)? 1:0);

  return 1;
}


static impulse_bool _p_window_key(XEvent  *event,
				       ROWS_widget widget)
{
  ROWS_child_event ev;

  /* Fix me: do keycode properly (ie convert to unicode) */
  ev.key.pressed = (event->type == KeyPress) ? 1 : 0;
  ev.key.key = event->xkey.keycode;
  ev.key.state = ROWS_X_get_state(event->xkey.state);

  ROWS_child_event_send(ROWS_child_cast(widget),
			ROWS_CHILD_KEY,
			&ev);
  
  return 1;
}


static impulse_bool _p_window_focus(XEvent  *event,
				    ROWS_widget widget)
{
  /* Fix me: do something useful */
  return 1;
}


static impulse_bool _p_window_client(XEvent  *event,
				     ROWS_widget widget)
{
  Atom deletewin, protocols;
  _p_ROWS_child_data  **priv;
  _p_ROWS_X_child_data *x_data;

  priv = (_p_ROWS_child_data **) ROWS_child_cast(widget);
  x_data = (_p_ROWS_X_child_data *)(*priv)->int_data;
  deletewin = XInternAtom(x_data->display,
			   "WM_DELETE_WINDOW",
			   1);
  protocols = XInternAtom(x_data->display,
			  "WM_PROTOCOLS",
			   1);
  
  if (event->xclient.type == protocols &&
      event->xclient.data.l[0] == deletewin)
  {
    return 1;
  }
  
  return 0;
}


/* ENd of file */

